@@ -28,7 +28,6 @@ from ast import (
     If,
     Import,
     ImportFrom,
-    Index,
     List,
     Load,
     LShift,
@@ -65,6 +64,9 @@ from copy import deepcopy
 from dataclasses import dataclass, field
 from typing import Any, ClassVar, cast, overload
 
+PreliminaryNameTypePair: typing.TypeAlias = tuple[Constant, "expr | None"]
+NameTypePair: typing.TypeAlias = tuple[Constant, expr]
+
 generator_names = (
     "typing.Generator",
     "collections.abc.Generator",
@@ -374,7 +376,7 @@ class AnnotationTransformer(NodeTransformer):
 
         if isinstance(node.op, BitOr):
             # If either branch of the BinOp has been transformed to `None`, it means
-            # that a type in the union was ignored, so the entire annotation should e
+            # that a type in the union was ignored, so the entire annotation should be
             # ignored
             if not hasattr(node, "left") or not hasattr(node, "right"):
                 return None
@@ -385,13 +387,12 @@ class AnnotationTransformer(NodeTransformer):
             elif self._memo.name_matches(node.right, *anytype_names):
                 return node.right
 
+            # Turn union types to typing.Union constructs on Python 3.9
             if sys.version_info < (3, 10):
                 union_name = self.transformer._get_import("typing", "Union")
                 return Subscript(
                     value=union_name,
-                    slice=Index(
-                        Tuple(elts=[node.left, node.right], ctx=Load()), ctx=Load()
-                    ),
+                    slice=Tuple(elts=[node.left, node.right], ctx=Load()),
                     ctx=Load(),
                 )
 
@@ -410,24 +411,18 @@ class AnnotationTransformer(NodeTransformer):
         # The subscript of typing(_extensions).Literal can be any arbitrary string, so
         # don't try to evaluate it as code
         if node.slice:
-            if isinstance(node.slice, Index):
-                # Python 3.8
-                slice_value = node.slice.value  # type: ignore[attr-defined]
-            else:
-                slice_value = node.slice
-
-            if isinstance(slice_value, Tuple):
+            if isinstance(node.slice, Tuple):
                 if self._memo.name_matches(node.value, *annotated_names):
                     # Only treat the first argument to typing.Annotated as a potential
                     # forward reference
                     items = cast(
                         typing.List[expr],
-                        [self.visit(slice_value.elts[0])] + slice_value.elts[1:],
+                        [self.visit(node.slice.elts[0])] + node.slice.elts[1:],
                     )
                 else:
                     items = cast(
                         typing.List[expr],
-                        [self.visit(item) for item in slice_value.elts],
+                        [self.visit(item) for item in node.slice.elts],
                     )
 
                 # If this is a Union and any of the items is Any, erase the entire
@@ -450,16 +445,18 @@ class AnnotationTransformer(NodeTransformer):
                     if item is None:
                         items[index] = self.transformer._get_import("typing", "Any")
 
-                slice_value.elts = items
+                node.slice.elts = items
             else:
                 self.generic_visit(node)
 
                 # If the transformer erased the slice entirely, just return the node
                 # value without the subscript (unless it's Optional, in which case erase
                 # the node entirely
-                if self._memo.name_matches(node.value, "typing.Optional"):
+                if self._memo.name_matches(
+                    node.value, "typing.Optional"
+                ) and not hasattr(node, "slice"):
                     return None
-                elif sys.version_info >= (3, 9) and not hasattr(node, "slice"):
+                if sys.version_info >= (3, 9) and not hasattr(node, "slice"):
                     return node.value
                 elif sys.version_info < (3, 9) and not hasattr(node.slice, "value"):
                     return node.value
@@ -470,12 +467,6 @@ class AnnotationTransformer(NodeTransformer):
         if self._memo.is_ignored_name(node):
             return None
 
-        if sys.version_info < (3, 9):
-            for typename, substitute in self.type_substitutions.items():
-                if self._memo.name_matches(node, typename):
-                    new_node = self.transformer._get_import(*substitute)
-                    return copy_location(new_node, node)
-
         return node
 
     def visit_Call(self, node: Call) -> Any:
@@ -546,18 +537,10 @@ class TypeguardTransformer(NodeTransformer):
                     return_annotation, *generator_names
                 ):
                     if isinstance(return_annotation, Subscript):
-                        annotation_slice = return_annotation.slice
-
-                        # Python < 3.9
-                        if isinstance(annotation_slice, Index):
-                            annotation_slice = (
-                                annotation_slice.value  # type: ignore[attr-defined]
-                            )
-
-                        if isinstance(annotation_slice, Tuple):
-                            items = annotation_slice.elts
+                        if isinstance(return_annotation.slice, Tuple):
+                            items = return_annotation.slice.elts
                         else:
-                            items = [annotation_slice]
+                            items = [return_annotation.slice]
 
                         if len(items) > 0:
                             new_memo.yield_annotation = self._convert_annotation(
@@ -589,12 +572,10 @@ class TypeguardTransformer(NodeTransformer):
         return memo.get_import(module, name)
 
     @overload
-    def _convert_annotation(self, annotation: None) -> None:
-        ...
+    def _convert_annotation(self, annotation: None) -> None: ...
 
     @overload
-    def _convert_annotation(self, annotation: expr) -> expr:
-        ...
+    def _convert_annotation(self, annotation: expr) -> expr: ...
 
     def _convert_annotation(self, annotation: expr | None) -> expr | None:
         if annotation is None:
@@ -723,7 +704,7 @@ class TypeguardTransformer(NodeTransformer):
                     else:
                         self.target_lineno = node.lineno
 
-                all_args = node.args.args + node.args.kwonlyargs + node.args.posonlyargs
+                all_args = node.args.posonlyargs + node.args.args + node.args.kwonlyargs
 
                 # Ensure that any type shadowed by the positional or keyword-only
                 # argument names are ignored in this function
@@ -748,21 +729,14 @@ class TypeguardTransformer(NodeTransformer):
                 if node.args.vararg:
                     annotation_ = self._convert_annotation(node.args.vararg.annotation)
                     if annotation_:
-                        if sys.version_info >= (3, 9):
-                            container = Name("tuple", ctx=Load())
-                        else:
-                            container = self._get_import("typing", "Tuple")
-
-                        subscript_slice: Tuple | Index = Tuple(
+                        container = Name("tuple", ctx=Load())
+                        subscript_slice = Tuple(
                             [
                                 annotation_,
                                 Constant(Ellipsis),
                             ],
                             ctx=Load(),
                         )
-                        if sys.version_info < (3, 9):
-                            subscript_slice = Index(subscript_slice, ctx=Load())
-
                         arg_annotations[node.args.vararg.arg] = Subscript(
                             container, subscript_slice, ctx=Load()
                         )
@@ -770,11 +744,7 @@ class TypeguardTransformer(NodeTransformer):
                 if node.args.kwarg:
                     annotation_ = self._convert_annotation(node.args.kwarg.annotation)
                     if annotation_:
-                        if sys.version_info >= (3, 9):
-                            container = Name("dict", ctx=Load())
-                        else:
-                            container = self._get_import("typing", "Dict")
-
+                        container = Name("dict", ctx=Load())
                         subscript_slice = Tuple(
                             [
                                 Name("str", ctx=Load()),
@@ -782,9 +752,6 @@ class TypeguardTransformer(NodeTransformer):
                             ],
                             ctx=Load(),
                         )
-                        if sys.version_info < (3, 9):
-                            subscript_slice = Index(subscript_slice, ctx=Load())
-
                         arg_annotations[node.args.kwarg.arg] = Subscript(
                             container, subscript_slice, ctx=Load()
                         )
@@ -803,7 +770,7 @@ class TypeguardTransformer(NodeTransformer):
                     ],
                 )
                 func_name = self._get_import(
-                    "typeguard._functions", "check_argument_types"
+                    "typeguard._functions", "check_argument_types_internal"
                 )
                 args = [
                     self._memo.joined_path,
@@ -826,7 +793,7 @@ class TypeguardTransformer(NodeTransformer):
                 )
             ):
                 func_name = self._get_import(
-                    "typeguard._functions", "check_return_type"
+                    "typeguard._functions", "check_return_type_internal"
                 )
                 return_node = Return(
                     Call(
@@ -863,19 +830,20 @@ class TypeguardTransformer(NodeTransformer):
                             isinstance(decorator, Name)
                             and decorator.id == "classmethod"
                         ):
+                            arglist = node.args.posonlyargs or node.args.args
                             memo_kwargs["self_type"] = Name(
-                                id=node.args.args[0].arg, ctx=Load()
+                                id=arglist[0].arg, ctx=Load()
                             )
                             break
                     else:
-                        if node.args.args:
+                        if arglist := node.args.posonlyargs or node.args.args:
                             if node.name == "__new__":
                                 memo_kwargs["self_type"] = Name(
-                                    id=node.args.args[0].arg, ctx=Load()
+                                    id=arglist[0].arg, ctx=Load()
                                 )
                             else:
                                 memo_kwargs["self_type"] = Attribute(
-                                    Name(id=node.args.args[0].arg, ctx=Load()),
+                                    Name(id=arglist[0].arg, ctx=Load()),
                                     "__class__",
                                     ctx=Load(),
                                 )
@@ -955,7 +923,9 @@ class TypeguardTransformer(NodeTransformer):
             and self._memo.should_instrument
             and not self._memo.is_ignored_name(self._memo.return_annotation)
         ):
-            func_name = self._get_import("typeguard._functions", "check_return_type")
+            func_name = self._get_import(
+                "typeguard._functions", "check_return_type_internal"
+            )
             old_node = node
             retval = old_node.value or Constant(None)
             node = Return(
@@ -1044,12 +1014,20 @@ class TypeguardTransformer(NodeTransformer):
                     func_name = self._get_import(
                         "typeguard._functions", "check_variable_assignment"
                     )
+                    targets_arg = List(
+                        [
+                            Tuple(
+                                [Constant(node.target.id), annotation],
+                                ctx=Load(),
+                            )
+                        ],
+                        ctx=Load(),
+                    )
                     node.value = Call(
                         func_name,
                         [
                             node.value,
-                            Constant(node.target.id),
-                            annotation,
+                            targets_arg,
                             self._memo.get_memo_name(),
                         ],
                         [],
@@ -1067,74 +1045,95 @@ class TypeguardTransformer(NodeTransformer):
 
         # Only instrument function-local assignments
         if isinstance(self._memo.node, (FunctionDef, AsyncFunctionDef)):
-            targets: list[dict[Constant, expr | None]] = []
+            preliminary_targets: list[
+                PreliminaryNameTypePair | list[PreliminaryNameTypePair]
+            ] = []
             check_required = False
-            for target in node.targets:
+            for node_target in node.targets:
                 elts: Sequence[expr]
-                if isinstance(target, Name):
-                    elts = [target]
-                elif isinstance(target, Tuple):
-                    elts = target.elts
+                if isinstance(node_target, Name):
+                    elts = [node_target]
+                    single_target = True
+                elif isinstance(node_target, Tuple):
+                    elts = node_target.elts
+                    single_target = False
                 else:
                     continue
 
-                annotations_: dict[Constant, expr | None] = {}
+                annotations_: list[PreliminaryNameTypePair] = []
                 for exp in elts:
                     prefix = ""
                     if isinstance(exp, Starred):
                         exp = exp.value
                         prefix = "*"
 
+                    path: list[str] = []
+                    while isinstance(exp, Attribute):
+                        path.insert(0, exp.attr)
+                        exp = exp.value
+
                     if isinstance(exp, Name):
-                        self._memo.ignored_names.add(exp.id)
-                        name = prefix + exp.id
-                        annotation = self._memo.variable_annotations.get(exp.id)
-                        if annotation:
-                            annotations_[Constant(name)] = annotation
+                        if not path:
+                            self._memo.ignored_names.add(exp.id)
+
+                        path.insert(0, exp.id)
+                        name = prefix + ".".join(path)
+                        if len(path) == 1 and (
+                            annotation := self._memo.variable_annotations.get(exp.id)
+                        ):
+                            annotations_.append((Constant(name), annotation))
                             check_required = True
                         else:
-                            annotations_[Constant(name)] = None
+                            annotations_.append((Constant(name), None))
 
-                targets.append(annotations_)
+                preliminary_targets.append(
+                    annotations_[0] if single_target else annotations_
+                )
 
             if check_required:
                 # Replace missing annotations with typing.Any
-                for item in targets:
-                    for key, expression in item.items():
-                        if expression is None:
-                            item[key] = self._get_import("typing", "Any")
+                targets: list[NameTypePair | list[NameTypePair]] = []
+                for items in preliminary_targets:
+                    if isinstance(items, list):
+                        target_list: list[NameTypePair] = []
+                        targets.append(target_list)
+                        for key, expression_or_none in items:
+                            expression = expression_or_none or self._get_import(
+                                "typing", "Any"
+                            )
+                            target_list.append((key, expression))
+                    else:
+                        key, expression_or_none = items
+                        expression = expression_or_none or self._get_import(
+                            "typing", "Any"
+                        )
+                        targets.append((key, expression))
 
-                if len(targets) == 1 and len(targets[0]) == 1:
-                    func_name = self._get_import(
-                        "typeguard._functions", "check_variable_assignment"
-                    )
-                    target_varname = next(iter(targets[0]))
-                    node.value = Call(
-                        func_name,
-                        [
-                            node.value,
-                            target_varname,
-                            targets[0][target_varname],
-                            self._memo.get_memo_name(),
-                        ],
-                        [],
-                    )
-                elif targets:
-                    func_name = self._get_import(
-                        "typeguard._functions", "check_multi_variable_assignment"
-                    )
-                    targets_arg = List(
-                        [
-                            Dict(keys=list(target), values=list(target.values()))
-                            for target in targets
-                        ],
-                        ctx=Load(),
-                    )
-                    node.value = Call(
-                        func_name,
-                        [node.value, targets_arg, self._memo.get_memo_name()],
-                        [],
-                    )
+                func_name = self._get_import(
+                    "typeguard._functions", "check_variable_assignment"
+                )
+                elements: list[expr] = []
+                for target in targets:
+                    if isinstance(target, list):
+                        elements.append(
+                            List(
+                                [
+                                    Tuple([name, ann], ctx=Load())
+                                    for name, ann in target
+                                ],
+                                ctx=Load(),
+                            )
+                        )
+                    else:
+                        name_constant, ann = target
+                        elements.append(Tuple([name_constant, ann], ctx=Load()))
+
+                targets_arg = List(elements, ctx=Load())
+                node.value = Call(
+                    func_name,
+                    [node.value, targets_arg, self._memo.get_memo_name()],
+                    [],
+                )
 
         return node
 
@@ -1160,8 +1159,20 @@ class TypeguardTransformer(NodeTransformer):
                 func_name,
                 [
                     node.value,
-                    Constant(node.target.id),
-                    annotation,
+                    List(
+                        [
+                            List(
+                                [
+                                    Tuple(
+                                        [Constant(node.target.id), annotation],
+                                        ctx=Load(),
+                                    )
+                                ],
+                                ctx=Load(),
+                            )
+                        ],
+                        ctx=Load(),
+                    ),
                     self._memo.get_memo_name(),
                 ],
                 [],
@@ -1195,12 +1206,15 @@ class TypeguardTransformer(NodeTransformer):
             operator_call = Call(
                 operator_func, [Name(node.target.id, ctx=Load()), node.value], []
             )
+            targets_arg = List(
+                [Tuple([Constant(node.target.id), annotation], ctx=Load())],
+                ctx=Load(),
+            )
             check_call = Call(
                 self._get_import("typeguard._functions", "check_variable_assignment"),
                 [
                     operator_call,
-                    Constant(node.target.id),
-                    annotation,
+                    targets_arg,
                     self._memo.get_memo_name(),
                 ],
                 [],
